import os
import base64
import requests
import subprocess

def message_to_indices(m): # from `sphincsplus/ref/fors.c`
    l = []
    offset = 0
    for i in range(17): # SPX_FORS_TREES
        l.append(0)
        for j in range(6): # SPX_FORS_HEIGHT
            l[i] ^= ((m[offset >> 3] >> (offset & 0x7)) & 0x1) << j
            offset += 1
    return l

def haraka(R, pk, m):
    mhash, node = subprocess.check_output(["./sphincsplus/ref/haraka_interface", R.hex(), pk.hex(), m.hex()]).decode().split('\n')[:-1]
    return bytes.fromhex(mhash.split('mhash = ')[1]), node.split('(tree_idx, leaf_idx) = ')[1]

def getpk():
    r = requests.get('https://quantumpyramids.web.ctfcompetition.com/publickey')
    return base64.b64decode(r.json()['publickey'])

def sign(m):
    r = requests.post('https://quantumpyramids.web.ctfcompetition.com/sign', json={'message': base64.b64encode(m).decode()})
    sig = base64.b64decode(r.json()['signature'])
    R = sig[:8]
    FORS_sig = sig[8:960]
    WOTS_sig = [sig[960:1104], sig[1104:1120], sig[1120:1264], sig[1264:1280]]
    return sig, R, FORS_sig, WOTS_sig

def verify(m,s):
    r = requests.post('https://quantumpyramids.web.ctfcompetition.com/verify', json={'message': base64.b64encode(m).decode(), 'signature': base64.b64encode(s).decode()})
    return r.text

pk = getpk()

# recover the secret key
## pick any node, e.g. (0, 2)
#sk = [dict() for _ in range(17)]
#it = 0
#stopper = False
#while not stopper:
#    m = os.urandom(32)
#    _, R, F, W = sign(m)
#    mhash, node = haraka(R, pk, m)
#    print(node)
#    if node != '(0, 2)':
#        continue
#    it+=1
#    print("#goods =", it)
#    temp = message_to_indices(mhash)
#    Fsplit = [(F[i:i+8], F[i+8:i+56]) for i in range(0, len(F), 56)]
#    for i,ki in enumerate(temp):
#        if ki not in sk[i].keys():
#            sk[i][ki] = Fsplit[i]
#        assert sk[i][ki] == Fsplit[i]
#    print(list(map(len, sk)))
#    stopper = all([le == 64 for le in map(len, sk)])
#with open('sk.dict', 'wb') as f:
#    f.write(str(sk).encode())

with open('sk.dict', 'rb') as f:
    sk = eval(f.read())

mb = b'opensesame'
node1 = False
node2 = False
while node1 != '(0, 2)' or node2 != '(0, 2)':
    m = os.urandom(10)
    _, R, F, W = sign(m)
    _, node1 = haraka(R, pk, m)
    _, node2 = haraka(R, pk, mb)

# "forge" the FORS for the real message
mhash, _ = haraka(R, pk, m)
indices = message_to_indices(mhash)
fors = b''
for i in range(len(sk)):
    xi, ai = (sk[i][indices[i]])
    fors += (xi+ai)
fakesig = R + fors + b''.join(W)

# forge the FORS for the target message
mhash2, _ = haraka(R, pk, mb)
indices2 = message_to_indices(mhash2)
fors2 = b''
for i in range(len(sk)):
    xi, ai = (sk[i][indices2[i]])
    fors2 += (xi+ai)
fakesig2 = R + fors2 + b''.join(W)

print(verify(m, fakesig))
print(verify(mb, fakesig2))
#{"status":"CTF{t00_m4ny_h4$h3s_f0r_t0d4y}"}
